from typing import Generic, TypeVar

from rainbowadn.hashing.hashpoint import HashPoint
from rainbowadn.hashing.nullability.nullablereference import NullableReference

__all__ = ('StageProtocol',)

HeaderType = TypeVar('HeaderType')
BaseStateType = TypeVar('BaseStateType')
StageType = TypeVar('StageType')


class StageProtocol(Generic[HeaderType, BaseStateType, StageType]):
    def verify_header(
            self,
            previous: NullableReference[BaseStateType],
            header: HashPoint[HeaderType],
            stage: HashPoint[StageType]
    ) -> bool:
        raise NotImplementedError

    def verify_stage(
            self,
            previous: HashPoint[StageType],
            stage: HashPoint[StageType]
    ) -> bool:
        raise NotImplementedError

    def verify_state(
            self,
            stage: HashPoint[StageType],
            state: HashPoint[BaseStateType]
    ) -> bool:
        raise NotImplementedError
