from typing import Generic, TypeVar

from rainbowadn.chain.stages.stage import StateStage
from rainbowadn.chain.states.stateprotocol import StateProtocol
from rainbowadn.hashing.hashpoint import HashPoint
from rainbowadn.hashing.nullability.nullablereference import NullableReference

__all__ = ('StageStateProtocol',)

HeaderType = TypeVar('HeaderType')
BaseStateType = TypeVar('BaseStateType')
StageType = TypeVar('StageType')


class StageStateProtocol(
    StateProtocol[HeaderType, StateStage[HeaderType, BaseStateType, StageType]],
    Generic[HeaderType, BaseStateType, StageType]
):
    def verify(
            self,
            previous: NullableReference[StateStage[HeaderType, BaseStateType, StageType]],
            header: HashPoint[HeaderType],
            state: HashPoint[StateStage[HeaderType, BaseStateType, StageType]]
    ) -> bool:
        assert isinstance(previous, NullableReference)
        assert isinstance(header, HashPoint)
        assert isinstance(state, HashPoint)
        return state.resolve().verify(previous, header)
