from typing import TypeVar

from rainbowadn.chain.states.stateprotocol import StateProtocol
from rainbowadn.hashing.hashpoint import HashPoint
from rainbowadn.hashing.nullability.notnull import NotNull
from rainbowadn.hashing.nullability.null import Null
from rainbowadn.hashing.nullability.nullablereference import NullableReference

__all__ = ('ActiveStateProtocol',)

HeaderType = TypeVar('HeaderType')
StateType = TypeVar('StateType')


class ActiveStateProtocol(StateProtocol[HeaderType, StateType]):
    def verify(
            self,
            previous: NullableReference[StateType],
            header: HashPoint[HeaderType],
            state: HashPoint[StateType]
    ) -> bool:
        assert isinstance(previous, NullableReference)
        assert isinstance(header, HashPoint)
        assert isinstance(state, HashPoint)
        assert state == self.derive(previous, header)
        return True

    def _initial_state(self) -> HashPoint[StateType]:
        raise NotImplementedError

    def _derive(
            self,
            previous: HashPoint[StateType],
            header: HashPoint[HeaderType],
    ) -> HashPoint[StateType]:
        raise NotImplementedError

    def derive(
            self,
            previous: NullableReference[StateType],
            header: HashPoint[HeaderType],
    ) -> HashPoint[StateType]:
        assert isinstance(previous, NullableReference)
        assert isinstance(header, HashPoint)
        if isinstance(previous.reference, Null):
            previous_state: HashPoint[StateType] = self._initial_state()
        elif isinstance(previous.reference, NotNull):
            previous_state: HashPoint[StateType] = previous.reference.value
        else:
            raise TypeError
        assert isinstance(previous_state, HashPoint)
        return self._derive(previous_state, header)
