from typing import Generic, Iterable, TypeVar

from rainbowadn.hashing.hash_point_format import hash_point_format, tabulate
from rainbowadn.hashing.hashpoint import HashPoint
from rainbowadn.hashing.hashresolver import HashResolver
from rainbowadn.hashing.rainbow_factory import RainbowFactory
from rainbowadn.hashing.recursivementionable import RecursiveMentionable
from rainbowadn.hashing.resolverorigin import ResolverOrigin

__all__ = ('Pair', 'PairFactory',)

E0Type = TypeVar('E0Type')
E1Type = TypeVar('E1Type')


class Pair(
    RecursiveMentionable,
    Generic[E0Type, E1Type]
):
    def __init__(self, element0: HashPoint[E0Type], element1: HashPoint[E1Type]):
        assert isinstance(element0, HashPoint)
        assert isinstance(element1, HashPoint)
        self.element0 = element0
        self.element1 = element1

    def points(self) -> Iterable[HashPoint]:
        return [self.element0, self.element1]

    def __bytes__(self):
        return bytes(self.element0) + bytes(self.element1)

    def __factory__(self) -> RainbowFactory['Pair[E0Type, E1Type]']:
        return PairFactory(self.element0.factory, self.element1.factory)

    def str(self, tab: int) -> str:
        assert isinstance(tab, int)
        return f'(pair)' \
               f'{tabulate(tab)}{hash_point_format(self.element0, tab)}' \
               f'{tabulate(tab)}{hash_point_format(self.element1, tab)}'


class PairFactory(
    RainbowFactory[Pair[E0Type, E1Type]],
    Generic[E0Type, E1Type]
):
    def __init__(
            self,
            e0_factory: RainbowFactory[E0Type],
            e1_factory: RainbowFactory[E1Type],
    ):
        assert isinstance(e0_factory, RainbowFactory)
        assert isinstance(e1_factory, RainbowFactory)
        self.e0_factory = e0_factory
        self.e1_factory = e1_factory

    def from_bytes(self, source: bytes, resolver: HashResolver) -> Pair[E0Type, E1Type]:
        assert isinstance(source, bytes)
        assert isinstance(resolver, HashResolver)
        return Pair(
            ResolverOrigin(self.e0_factory, source[:HashPoint.HASH_LENGTH], resolver).hash_point(),
            ResolverOrigin(self.e1_factory, source[HashPoint.HASH_LENGTH:], resolver).hash_point(),
        )
