from typing import TypeVar

from rainbowadn.data.atomic.integer import Integer
from rainbowadn.data.collection.trees.binary.binarytreebalancingprotocol import BinaryTreeBalancingProtocol
from rainbowadn.data.collection.trees.binary.binarytreecreationprotocol import BinaryTreeCreationProtocol
from rainbowadn.hashing.hashpoint import HashPoint

__all__ = ('AVLBTBP',)

ActiveKeyType = TypeVar('ActiveKeyType')
TreeType = TypeVar('TreeType')


class AVLBTBP(BinaryTreeBalancingProtocol[ActiveKeyType, Integer, TreeType]):
    def empty_metadata(self) -> HashPoint[Integer]:
        return HashPoint.of(Integer(0))

    def metadata(
            self,
            treel: TreeType,
            treer: TreeType,
            key: HashPoint[ActiveKeyType],
            protocol: BinaryTreeCreationProtocol[ActiveKeyType, Integer, TreeType]
    ) -> HashPoint[Integer]:
        return HashPoint.of(
            Integer(1 + max(self.height(treel, protocol), self.height(treer, protocol)))
        )

    @classmethod
    def height(
            cls,
            tree: TreeType,
            protocol: BinaryTreeCreationProtocol[ActiveKeyType, Integer, TreeType]
    ) -> int:
        if protocol.splittable(tree):
            _, _, metadata, _ = protocol.split(tree)
            return metadata.resolve().integer
        else:
            return 0

    def balance(
            self,
            tree: TreeType,
            protocol: BinaryTreeCreationProtocol[ActiveKeyType, Integer, TreeType]
    ) -> TreeType:
        if protocol.splittable(tree):
            treel, key, _, treer = protocol.split(tree)
            delta = self.height(treel, protocol) - self.height(treer, protocol)
            assert isinstance(delta, int)
            if delta < -1:
                assert protocol.splittable(treer)
                treerl, keyr, _, treerr = protocol.split(treer)
                if self.height(treerl, protocol) > self.height(treerr, protocol):
                    assert protocol.splittable(treerl)
                    treerll, keyrl, _, treerlr = protocol.split(treerl)
                    return protocol.tree(
                        protocol.tree(treel, treerll, key),
                        protocol.tree(treerlr, treerr, keyr),
                        keyrl
                    )
                else:
                    return protocol.tree(
                        protocol.tree(treel, treerl, key),
                        treerr,
                        keyr
                    )
            elif delta > 1:
                assert protocol.splittable(treel)
                treell, keyl, _, treelr = protocol.split(treel)
                if self.height(treelr, protocol) > self.height(treell, protocol):
                    assert protocol.splittable(treelr)
                    treelrl, keylr, _, treelrr = protocol.split(treelr)
                    return protocol.tree(
                        protocol.tree(treell, treelrl, keyl),
                        protocol.tree(treelrr, treer, key),
                        keylr
                    )
                else:
                    return protocol.tree(
                        treell,
                        protocol.tree(treelr, treer, key),
                        keyl
                    )
            else:
                return tree
        else:
            return tree
