from typing import Generic, TypeVar

from rainbowadn.data.collection.trees.binary.binarytreebalancingprotocol import BinaryTreeBalancingProtocol
from rainbowadn.data.collection.trees.binary.binarytreecreationprotocol import BinaryTreeCreationProtocol
from rainbowadn.hashing.hashpoint import HashPoint

__all__ = ('BalancedTreeCreationProtocol',)

TreeType = TypeVar('TreeType')
ActiveKeyType = TypeVar('ActiveKeyType')
MetaDataType = TypeVar('MetaDataType')


class BalancedTreeCreationProtocol(
    BinaryTreeCreationProtocol[ActiveKeyType, MetaDataType, TreeType],
    Generic[ActiveKeyType, MetaDataType, TreeType]
):
    def __init__(
            self,
            protocol: BinaryTreeBalancingProtocol[ActiveKeyType, MetaDataType, TreeType]
    ):
        assert isinstance(protocol, BinaryTreeBalancingProtocol)
        self.protocol = protocol

    def splittable(self, tree: TreeType) -> bool:
        raise NotImplementedError

    def split(self, tree: TreeType) -> tuple[TreeType, HashPoint[ActiveKeyType], HashPoint[MetaDataType], TreeType]:
        raise NotImplementedError

    def _tree(self, treel: TreeType, treer: TreeType, key: HashPoint[ActiveKeyType]) -> TreeType:
        raise NotImplementedError

    def tree(self, treel: TreeType, treer: TreeType, key: HashPoint[ActiveKeyType]) -> TreeType:
        assert isinstance(key, HashPoint)
        return self.protocol.balance(
            self._tree(
                self.protocol.balance(treel, self),
                self.protocol.balance(treer, self),
                key
            ),
            self
        )
