from typing import Generic, TypeVar

from rainbowadn.data.collection.collection_interface.querycollectioninterface import QueryCollectionInterface
from rainbowadn.data.collection.trees.binary.binarytree import BinaryTree
from rainbowadn.data.collection.trees.comparison.comparator import Comparator, Comparison, Equal, Left, Right
from rainbowadn.hashing.hashpoint import HashPoint
from rainbowadn.hashing.nullability.notnull import NotNull
from rainbowadn.hashing.nullability.null import Null
from rainbowadn.hashing.nullability.nullable import Nullable
from rainbowadn.hashing.nullability.nullablereference import NullableReference

__all__ = ('QueryBinaryTree',)

KeyType = TypeVar('KeyType')


class QueryBinaryTree(QueryCollectionInterface[KeyType], Generic[KeyType]):
    def __init__(
            self, comparator: Comparator[KeyType], reference: NullableReference[BinaryTree[KeyType]]
    ):
        assert isinstance(comparator, Comparator)
        assert isinstance(reference, NullableReference)
        self.comparator = comparator
        self.reference = reference

    def query(self, key: HashPoint[KeyType]) -> Nullable[HashPoint[KeyType]]:
        assert isinstance(key, HashPoint)
        reference: Nullable[HashPoint[BinaryTree[KeyType]]] = self.reference.reference
        assert isinstance(reference, Nullable)
        if isinstance(reference, Null):
            return Null()
        elif isinstance(reference, NotNull):
            hash_point: HashPoint[BinaryTree[KeyType]] = reference.value
            assert isinstance(hash_point, HashPoint)
            tree: BinaryTree[KeyType] = hash_point.resolve()
            assert isinstance(tree, BinaryTree)
            comparison: Comparison = self.comparator.compare(tree.key, key)
            assert isinstance(comparison, Comparison)
            if isinstance(comparison, Equal):
                return NotNull(tree.key)
            elif isinstance(comparison, Left):
                return QueryBinaryTree(self.comparator, tree.treel).query(key)
            elif isinstance(comparison, Right):
                return QueryBinaryTree(self.comparator, tree.treer).query(key)
            else:
                raise TypeError
        else:
            raise TypeError
