import hashlib
from typing import Generic, TypeVar

from rainbowadn.hashing.hashmentionable import HashMentionable
from rainbowadn.hashing.localorigin import LocalOrigin
from rainbowadn.hashing.origin import Origin
from rainbowadn.hashing.rainbow_factory import RainbowFactory

__all__ = ('HashPoint',)

HashMentioned = TypeVar('HashMentioned')
ReHashMentioned = TypeVar('ReHashMentioned')


def _hash(source: bytes) -> bytes:
    assert isinstance(source, bytes)
    return hashlib.sha256(source).digest()


class HashPoint(Generic[HashMentioned]):
    def __init__(
            self,
            factory: RainbowFactory[HashMentioned],
            point: bytes,
            origin: Origin[HashMentioned]
    ):
        assert isinstance(factory, RainbowFactory)
        assert isinstance(point, bytes)
        assert isinstance(origin, Origin)
        assert len(point) == self.HASH_LENGTH
        self.factory = factory
        self.point = point
        self.origin = origin

    def __bytes__(self):
        return self.point

    HASH_LENGTH = len(_hash(b''))
    NULL_HASH = b'\0' * HASH_LENGTH

    @classmethod
    def hash(cls, source: bytes) -> bytes:
        assert isinstance(source, bytes)
        return _hash(source)

    @classmethod
    def bytes_of_mentioned(cls, mentioned: HashMentionable):
        assert isinstance(mentioned, HashMentionable)
        topology_hash: bytes = mentioned.__topology_hash__()
        assert isinstance(topology_hash, bytes)
        assert len(topology_hash) == cls.HASH_LENGTH
        return topology_hash + bytes(mentioned)

    @classmethod
    def of(cls, mentioned: HashMentioned) -> 'HashPoint[HashMentioned]':
        assert isinstance(mentioned, HashMentionable)
        return cls(
            mentioned.__factory__(), cls.hash(cls.bytes_of_mentioned(mentioned)), LocalOrigin(mentioned)
        )

    def resolve(self) -> HashMentioned:
        resolved = self.origin.resolve()
        assert isinstance(resolved, HashMentionable)
        assert self.point == self.hash(self.bytes_of_mentioned(resolved))
        return resolved

    def __eq__(self, other):
        if isinstance(other, HashPoint):
            return self.point == other.point
        else:
            return NotImplemented
