from typing import Generic, TypeVar

from rainbowadn.hashing.hashpoint import HashPoint
from rainbowadn.hashing.hashresolver import HashResolver
from rainbowadn.hashing.metaorigin import MetaOrigin
from rainbowadn.hashing.origin import Origin
from rainbowadn.hashing.rainbow_factory import RainbowFactory
from rainbowadn.hashing.resolverorigin import ResolverOrigin

__all__ = ('ResolverMetaOrigin',)

OriginType = TypeVar('OriginType')


class ResolverMetaOrigin(MetaOrigin[OriginType], Generic[OriginType]):
    def __init__(self, resolver: HashResolver):
        assert isinstance(resolver, HashResolver)
        self.resolver = resolver

    def origin(self, factory: RainbowFactory[OriginType], point: bytes) -> Origin[OriginType]:
        assert isinstance(factory, RainbowFactory)
        assert isinstance(point, bytes)
        assert len(point) == HashPoint.HASH_LENGTH
        return ResolverOrigin(factory, point, self.resolver)

    def migrate(self, hash_point: HashPoint[OriginType]) -> HashPoint[OriginType]:
        assert isinstance(hash_point, HashPoint)
        return self.hash_point(hash_point.factory, hash_point.point)
