from typing import Generic, TypeVar

from rainbowadn.chain.abstractreductionchainmetafactory import AbstractReductionChainMetaFactory
from rainbowadn.chain.chaincollectioninterface import ChainCollectionInterface
from rainbowadn.data.collection.stack.stack import Stack, StackFactory
from rainbowadn.hashing.hashpoint import HashPoint
from rainbowadn.hashing.nullability.nullablereference import NullableReference, NullableReferenceFactory
from rainbowadn.v13.bankprotocol import BankProtocol
from rainbowadn.v13.bankstate import BankState
from rainbowadn.v13.transaction import Transaction

__all__ = ('BankChain',)

BlockType = TypeVar('BlockType')


class BankChain(Generic[BlockType]):
    def __init__(
            self,
            chain: ChainCollectionInterface[
                BlockType,
                NullableReference[Stack[Transaction]],
                BankState,
            ]
    ):
        assert isinstance(chain, ChainCollectionInterface)
        self.chain = chain
        self.reference = self.chain.reference
        assert isinstance(self.reference, NullableReference)

    @classmethod
    def empty(
            cls,
            factory: AbstractReductionChainMetaFactory[BlockType, NullableReference[Stack[Transaction]], BankState]
    ) -> 'BankChain[BlockType]':
        assert isinstance(factory, AbstractReductionChainMetaFactory)
        return cls(
            factory.factory(
                NullableReferenceFactory(StackFactory(Transaction.factory()).loose()).loose(),
                BankState.factory(),
                BankProtocol(),
            ).empty()
        )

    @classmethod
    def from_reference(
            cls,
            factory: AbstractReductionChainMetaFactory[BlockType, NullableReference[Stack[Transaction]], BankState],
            reference: NullableReference[
                BlockType
            ]
    ) -> 'BankChain[BlockType]':
        assert isinstance(factory, AbstractReductionChainMetaFactory)
        assert isinstance(reference, NullableReference)
        return cls(
            factory.factory(
                NullableReferenceFactory(StackFactory(Transaction.factory()).loose()).loose(),
                BankState.factory(),
                BankProtocol(),
            ).from_reference(
                reference
            )
        )

    def add(self, stack: NullableReference[Stack[Transaction]]) -> 'BankChain[BlockType]':
        assert isinstance(stack, NullableReference)
        return BankChain(
            self.chain.add(HashPoint.of(stack))
        )

    def adds(self, transactions: list[Transaction]) -> 'BankChain[BlockType]':
        assert isinstance(transactions, list)
        return self.add(
            Stack.off(
                Transaction.factory(),
                reversed(transactions)
            )
        )

    def verify(self) -> bool:
        return self.chain.verify()

    def __str__(self):
        return self.reference.str(0)
