from typing import TypeVar

from rainbowadn.chain.states.activestateprotocol import ActiveStateProtocol
from rainbowadn.hashing.hashpoint import HashPoint
from rainbowadn.wrisbt.wrisbtindex import WrisbtIndex
from rainbowadn.wrisbt.wrisbtparametres import WrisbtParametres
from rainbowadn.wrisbt.wrisbtroot import WrisbtRoot

__all__ = ('WrisbtProtocol',)

TargetType = TypeVar('TargetType')


class WrisbtProtocol(ActiveStateProtocol[TargetType, WrisbtIndex]):
    def __init__(self, keymin: int):
        assert isinstance(keymin, int)
        assert keymin >= 2
        self.keymin = keymin

    def _initial_state(self) -> HashPoint[WrisbtIndex]:
        return HashPoint.of(
            WrisbtIndex(
                HashPoint.of(WrisbtRoot.empty(WrisbtParametres(self.keymin, HashPoint.HASH_LENGTH))),
                HashPoint.of(WrisbtRoot.empty(WrisbtParametres(self.keymin, HashPoint.HASH_LENGTH))),
                self.keymin
            )
        )

    def _derive(
            self,
            previous: HashPoint[WrisbtIndex],
            header: HashPoint[TargetType]
    ) -> HashPoint[WrisbtIndex]:
        assert isinstance(previous, HashPoint)
        assert isinstance(header, HashPoint)
        index: WrisbtIndex = previous.resolve()
        assert isinstance(index, WrisbtIndex)

        empty: WrisbtRoot = WrisbtRoot.empty(WrisbtParametres(self.keymin, HashPoint.HASH_LENGTH))
        assert isinstance(empty, WrisbtRoot)
        total: WrisbtRoot = index.total.resolve()
        assert isinstance(total, WrisbtRoot)

        return HashPoint.of(
            WrisbtIndex(
                HashPoint.of(total.index(header, empty)),
                HashPoint.of(empty.index(header, total)),
                index.keymin
            )
        )
